Program Test3D;

{$hints off}
{$apptype gui}
{$codepage utf8}

uses
  Windows,
  GlExt,
  Glu,
  Gl;

const
  VK_W = $057;
  VK_S = $053;
  VK_A = $041;
  VK_D = $044;

type
  TVector3D = record
     x, y, z: Double;
  end;

  TCamera = record
     Pos   : TVector3D;
     View  : TVector3D;
     Up    : TVector3D;
  end;

var
  Msg: TMsg;
  Camera: TCamera;
  WindowClass: TWndClassEx;
  hWindow, hDC, hRC: THandle;
  AppWidth, AppHeight: Integer;

procedure DrawGrid();
var
  i: Double;
begin
  i := -50;
  while i <= 100 do
  begin
    i := i + 5;
    glBegin(GL_LINES);
	glVertex3f(-50, 0, i);
	glVertex3f(100, 0, i);
	glVertex3f(i, 0,-50);
	glVertex3f(i, 0, 100);
    glEnd();
  end;
end;

procedure RenderScene(hDC: THandle);
begin
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
  glLoadIdentity();

  gluLookAt(Camera.Pos.x,  Camera.Pos.y,  Camera.Pos.z,
            Camera.View.x, Camera.View.y, Camera.View.z,
            Camera.Up.x,   Camera.Up.y,   Camera.Up.z);

  DrawGrid();
  glTranslatef(0.0, 30.0, 0.0);
  DrawGrid();;

  SwapBuffers(hDC);
end;

function NormalizeVector(Vector: TVector3D): TVector3D;
var
  Length: Double;
begin
  Length := sqrt((Vector.x * Vector.x) + (Vector.y * Vector.y) + (Vector.z * Vector.z));

  Result.x := Vector.x / Length;
  Result.y := Vector.y / Length;
  Result.z := Vector.z / Length;
end;

procedure RotateCamera(Speed: Double);
var
  Vector: TVector3D;
begin
  Vector.x := Camera.View.x - Camera.Pos.x;
  Vector.z := Camera.View.z - Camera.Pos.z;

  Camera.View.z := Camera.Pos.z + sin(Speed) * Vector.x + cos(Speed) * Vector.z;
  Camera.View.x := Camera.Pos.x + cos(Speed) * Vector.x - sin(Speed) * Vector.z;
end;

procedure MoveCamera(Speed: Double);
var
  Vector: TVector3D;
begin
  Vector.x := Camera.View.x - Camera.Pos.x;
  Vector.y := 0;
  Vector.z := Camera.View.z - Camera.Pos.z;

  Vector := NormalizeVector(Vector);

  Camera.Pos.x  := Camera.Pos.x  + Vector.x * Speed;
  Camera.Pos.z  := Camera.Pos.z  + Vector.z * Speed;
  Camera.View.x := Camera.View.x + Vector.x * Speed;
  Camera.View.z := Camera.View.z + Vector.z * Speed;
end;

procedure MoveMouse(xMouse, yMouse: Integer);
var
  xAngle, yAngle: Double;
begin
  xAngle := Double((AppWidth  shr 1) - xMouse) * 0.0005;
  yAngle := Double((AppHeight shr 1) - yMouse) * 0.0010;

  if Camera.View.y > 8.0 then
     Camera.View.y := 8.0;
  if Camera.View.y < -3.0 then
     Camera.View.y := -3.0;

  Camera.View.y := Camera.View.y + yAngle;
  RotateCamera(-xAngle);

  SetCursorPos(AppWidth shr 1, AppHeight shr 1);
  RenderScene(hDC);
end;

procedure KeyboardInput(Key: Integer);
begin
  case Key of
    VK_W: MoveCamera(0.5);
    VK_S: MoveCamera(-0.5);
    VK_D: RotateCamera(0.025);
    VK_A: RotateCamera(-0.025);
    VK_ESCAPE: PostQuitMessage(0);
  end;
end;

procedure ResizeScene(nWidth, nHeight: Integer);
begin
  AppWidth  := nWidth;
  AppHeight := nHeight;

  glViewport(0, 0, nWidth, nHeight);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0, nWidth / nHeight, 0.1, 0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
end;

procedure EnableOpenGL(hWindow: THandle; var hDC, hRC: THandle);
var
  PixelFormat: TPixelFormatDescriptor;
begin
  ZeroMemory(@PixelFormat, (SizeOf(PIXELFORMATDESCRIPTOR)));

  PixelFormat.nSize      := SizeOf(PIXELFORMATDESCRIPTOR);
  PixelFormat.nVersion   := 1;
  PixelFormat.dwFlags    := PFD_DRAW_TO_WINDOW or
                            PFD_SUPPORT_OPENGL or
                            PFD_DOUBLEBUFFER;
  PixelFormat.iPixelType := PFD_TYPE_RGBA;
  PixelFormat.cColorBits := 24;
  PixelFormat.cDepthBits := 16;
  PixelFormat.iLayerType := PFD_MAIN_PLANE;

  hDC := GetDC(hWindow);
  if hDC = 0 then
     Exit;
  if not SetPixelFormat(hDC, (ChoosePixelFormat(hDC, @PixelFormat)), @PixelFormat) then
     Exit;
  hRC := wglCreateContext(hDC);
  if hRC = 0 then
     Exit;
  if not wglMakeCurrent(hDC, hRC) then
     Exit;

  ZeroMemory(@Camera, (SizeOf(TCamera)));

  Camera.Pos.y  := 2.5;
  Camera.Pos.z  := 5.0;
  Camera.View.y := 2.5;
  Camera.Up.y   := 1.0;
end;

function WindowProc(hWindow: THandle; uMsg: Cardinal; wParam: WParam; lParam: LParam): LResult; stdcall;
begin
  case uMsg of
    WM_SIZE      : ResizeScene(LOWORD(lParam), (HIWORD(lParam)));
    WM_PAINT     : RenderScene(hDC);
    WM_CREATE    : EnableOpenGL(hWindow, hDC, hRC);
    WM_KEYDOWN   : KeyboardInput(wParam);
    WM_MOUSEMOVE : MoveMouse(LOWORD(lParam), (HIWORD(lParam)));
  end;
  Result := DefWindowProc(hWindow, uMsg, wParam, lParam);
end;

begin
  WindowClass.cbSize        := SizeOf(WindowClass);
  WindowClass.style         := CS_OWNDC;
  WindowClass.lpfnWndProc   := @WindowProc;
  WindowClass.cbClsExtra    := 0;
  WindowClass.cbWndExtra    := 0;
  WindowClass.hInstance     := hInstance;
  WindowClass.hIcon         := LoadIcon(0, IDI_APPLICATION);
  WindowClass.hCursor       := LoadCursor(0, IDC_ARROW);
  WindowClass.hbrBackground := 0;
  WindowClass.lpszMenuName  := nil;
  WindowClass.lpszClassName := 'GLClass';
  WindowClass.hIconSm       := LoadIcon(0, IDI_APPLICATION);

  if RegisterClassEx(@WindowClass) = 0 then
  begin
    MessageBox(0, 'Невозможно зарегистрировать класс окна.', 'Шайссе!', MB_ICONERROR);
    Exit;
  end;

  hWindow := CreateWindowEx(0,
                            'GLClass',
                            'OpenGL 3D',
                            WS_POPUPWINDOW,
                            0,
                            0,
                            800,
                            600,
                            0,
                            0,
                            hInstance,
                            nil);

  if hWindow = 0 then
  begin
    MessageBox(0, 'Невозможно создать окно.', 'Шайссе!', MB_ICONERROR);
    Exit;
  end;

  ShowWindow(hWindow, SW_MAXIMIZE);
  ShowCursor(false);

  ZeroMemory(@Msg, (SizeOf(TMsg)));

  while (Msg.message <> WM_QUIT) do
  begin
    if PeekMessage(@Msg, 0, 0, 0, PM_REMOVE) then
    begin
      TranslateMessage(@Msg);
      DispatchMessage(@Msg);
    end
      else SendMessage(hWindow, WM_PAINT, 0, 0);
  end;
  Halt(Msg.wParam);
end.
